#include <stdio.h>
#include <string.h>

//#define KORGM   // for is40
//#define KORGI	  // for i30

static int srckorgm = 0;
static const char* korgmname = 0;
static char* korg9prefix = 0;
static char korg9name[128];
static FILE* korgm = 0;
static FILE* korg9 = 0;

#define KORGM_FILEID  "KORGM"
#define KORGI_FILEID  "KORGI"

#define KORGM_NAMESIZE  16

static void writeword2(unsigned v)
{
   int b = (v & 0xff);
   fputc(b, korg9);
   v >>= 8;
   b = (v & 0xff);
   fputc(b, korg9);
}

static long readlong(FILE* input)
{
  long v = 0;
  for (int i = 0; i < 32; i+=8)
  {
    int c = fgetc(input);
    v += ((long)c)<< i;
  }
  return v;
}

static int readword(FILE* input)
{
  int v = 0;
  for (int i = 0; i < 16; i+=8)
  {
    int c = fgetc(input);
    v += c << i;
  }
  return v;
}

static void writeword(int v, FILE* output)
{
  fputc(v & 0xff, output);
  fputc((v >>8) & 0xff, output);
}

static long copybytes(FILE* input, FILE* output, long size)
{
  long n = 0;
  while (size-- > 0)
  {
   fputc(fgetc(input), output);
   n++;
  }
  return n;
}

static void copystyle(int idx, char* name)
{
  fseek(korgm, 16, SEEK_SET);
  long stylepos = readword(korgm);
  unsigned stylelen = 0;
  fseek(korgm, 6, SEEK_CUR);
  for (int i = 0; i <= idx; i++)
  {
    stylelen = readword(korgm);
    readword(korgm); // id
    if (i == idx)
      break;
    stylepos += stylelen;
  }
  fseek(korgm, stylepos+18, SEEK_SET);
  stylelen -= 18;

  //unsigned newstylepos = (unsigned)ftell (korg9);
  fwrite(name, 10, 1, korg9);
  stylelen -= copybytes(korgm, korg9, 28);
  for (int trk = 0; trk < 6; trk++)
  {
    int prgm = fgetc(korgm);
    int bankm = fgetc(korgm);
    int prg9 = prgm;
    int bank9 = bankm;
    if (bankm == 4)
    {
      switch(prgm)
      {
      case 3: prg9 = 1; break;
      case 5: prg9 = 2; break;
      case 7: prg9 = 3; break;
      case 8: prg9 = 4; break;
      case 10: prg9 = 5; break;
      case 14: case 32: prg9 = 6; break;
      case 15: case 33: prg9 = 7; break;
      }
      bank9 = 2;
    }
    fputc(prg9, korg9);
    fputc(bank9, korg9);
    stylelen-= 2;
    stylelen -= copybytes(korgm, korg9, 2);
  }
  stylelen -= copybytes(korgm, korg9, 472);
  for (i = 0; i< 317; i++)
  {
     unsigned ofs = readword(korgm) - 8;
     writeword(ofs, korg9);
     stylelen-=2;
  }
  while (stylelen-- > 0)
   fputc(fgetc(korgm), korg9);
}

static void writelong3(long v)
{
   v = (v + 15) & 0xFFFFFF0L;
   int b = (v & 0xff);
   fputc(b, korg9);
   v >>= 8;
   fputc(0, korg9);
   b = (v & 0xff);
   fputc(b, korg9);
   v >>= 8;
   // v should be 0
}



static int convertkorgmto9(int idx)
{
  korgm = fopen(korgmname, "rb");
  if (!korgm)
  {
    perror(korgmname);
    return 0;
  }
  char id[16];
  fread(id, 1, 16, korgm);
  if (strncmp(id, KORGM_FILEID, 5) == 0)
    srckorgm = 1;
  else if (strncmp(id, KORGI_FILEID, 5) == 0)
    srckorgm = 0;
  else
  {
    fprintf(stderr, "%s is not a %s or %s file\n", korgmname, KORGM_FILEID, KORGI_FILEID);
    fclose(korgm);
    return 0;
  }

  fseek(korgm, 16, SEEK_SET);
  long firststylepos = readword(korgm);
  long stylepos = firststylepos;
  unsigned stylelen = 0;
  fseek(korgm, 6, SEEK_CUR);
  for (int i = 0; i <= idx && ftell(korgm) < firststylepos; i++)
  {
    stylelen = readword(korgm);
    readword(korgm); // id
    if (i == idx)
      break;
    stylepos += stylelen;
  }
  if (stylelen== 1176) // empty
    stylelen = 0;
  fseek(korgm, stylepos, SEEK_SET);
  char name[KORGM_NAMESIZE+1];
  if (fread(name, KORGM_NAMESIZE, 1, korgm) != 1)
    return 0;
  name[KORGM_NAMESIZE] = 0;
  readword(korgm); // ?

  sprintf(korg9name, "%s%d.sty", korg9prefix, idx);

  if (stylelen == 0)
  {
    fprintf(stderr, "style %s is empty\r", name);
    fclose(korgm);
    return 1;
  }
  printf("copying style %.18s to %s\n", name, korg9name);

  korg9 = fopen(korg9name, "wb");
  if (!korg9)
  {
    perror(korg9name);
    return 0;
  }

  fwrite("KORG9\0\x04\0\0\0\0\0\0\0\0\0", 16, 1, korg9);
  fwrite("\x20\0\0\0\x28\0\0\0\x48\0\0\0\x18\0\0\0", 16, 1, korg9);

  fprintf(korg9, "%10.10sempty     empty     empty     ", name);

  long pagepos = 0x20;
  writelong3(pagepos); fputc(4, korg9);

  long updatelenpos = ftell(korg9);
  writeword2(0xffffU); // length of style (later)
  writelong3(0); fputc(4, korg9);
  writeword2(0x490); // empty style
  writelong3(0); fputc(4, korg9);
  writeword2(0x490); // empty style
  writelong3(0); fputc(4, korg9);
  writeword2(0x490); // empty style

  long stylestartpos = ftell(korg9);

  copystyle(idx, name);

  long styleendpos = ftell(korg9);
  fseek(korg9, updatelenpos, SEEK_SET);
  stylelen = (unsigned)(styleendpos-stylestartpos);
  writeword2(stylelen);
  pagepos += stylelen;
  writelong3(pagepos); fputc(4, korg9);
  unsigned emptystylelen = 0x490;
  writeword2(emptystylelen);
  pagepos += emptystylelen;
  writelong3(pagepos); fputc(4, korg9);
  writeword2(emptystylelen);
  pagepos += emptystylelen;
  writelong3(pagepos); fputc(4, korg9);
  writeword2(emptystylelen);
  pagepos += emptystylelen;

  fseek(korg9, styleendpos, SEEK_SET);

  for (int j = 0; j < 3; j++)
  {
    fwrite("empty     ", 10, 1, korg9);
    for (i = 0; i < 0x144; i++)
      fputc(0, korg9);
    for (i = 0; i < 100; i++)
    {
      fputc(0x13, korg9);
      fputc(4, korg9);
    }
    for (i = 0; i < 317; i++)
    {
      writeword(0x490, korg9);
    }
  }
  fclose(korgm);
  fclose(korg9);
  return 1;
}

#ifdef NOMAIN
int main_stymito9(int argc, char** argv)
#else
int main(int argc, char** argv)
#endif
{
  argc--; argv++;

  if (argc != 2)
  {
#ifdef KORGM
    printf("usage: stymto9 korgstym.sty outputprefix\n");
    printf("korgstym.sty\ta korg format M style (e.g. for is40,is50 keyboard)\n");
#elif defined(KORGI)
    printf("usage: styito9 korgstyi.sty outputprefix\n");
    printf("korgstyi.sty\ta korg format I style (e.g. for i30 keyboard)\n");
#endif
    printf("outputprefix\tconverter will write files outputprefix1.sty, outputprefix2.sty ...\n");
    return 1;
  }
  korgmname = *argv++; argc--;
  korg9prefix = *argv++; argc--;

  for (int i = 0; i < 100; i++)
    if (!convertkorgmto9(i))
      break;
  return i > 0 ? 0 : 1;
}
